#include "pixmap.hh"
#include "conf.hh"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <GL/gl.h>
#include "glext.hh"
#include "util.hh"
extern "C" {
#include <jpeglib.h>
}

class JPEGPixmap : public Pixmap {
    struct jpeg_decompress_struct cinfo;
    FILE *ImageFile;
    int bytespersample;
    struct jpeg_error_mgr jerr;
  public:
    JPEGPixmap(const char *fn, bool read_whole);
    ~JPEGPixmap();
    bool read_scanline(char *data);
    void read_scanlines(unsigned char **data[3]);
};

class JPEGPixmapFactory : public PixmapFactory {
    bool read_whole;
  public:
    JPEGPixmapFactory(bool read_whole);
    Pixmap *create(const char *fn);
    bool accept_extension(const char *ext);
    const char *get_name();
};

const char *JPEGPixmapFactory::get_name()
{
  return "JPEG Image";
}

JPEGPixmap::~JPEGPixmap()
{
  if(image_valid) {
    if(!use_data) {
      jpeg_abort_decompress(&cinfo);
      fclose(ImageFile);
      jpeg_destroy_decompress(&cinfo);
    }
  }
}

bool JPEGPixmapFactory::accept_extension(const char *ext)
{
  return !strcasecmp(ext, "jpg")||!strcasecmp(ext, "jpeg");
}

JPEGPixmapFactory::JPEGPixmapFactory(bool read_whole) : read_whole(read_whole)
{
}

Pixmap *JPEGPixmapFactory::create(const char *fn)
{
  return new JPEGPixmap(fn, read_whole);
}

JPEGPixmapFactory jpeg_pixmap_factory(true);

JPEGPixmap::JPEGPixmap(const char *fn, bool read_whole)
{
  w=-1, h=-1, data=NULL;
  cinfo.err=jpeg_std_error(&jerr);
  jpeg_create_decompress(&cinfo);
  ImageFile=fopen(fn, "r");
  flip=false;
  if(!ImageFile)
    return;
  w=cinfo.output_width;
  h=cinfo.output_height;
  jpeg_stdio_src(&cinfo, ImageFile);
  jpeg_read_header(&cinfo, TRUE);
  int in_colorspace;
  printf("vsampf=%d, hsampf=%d, w=%d, h=%d\n", cinfo.comp_info[0].v_samp_factor, cinfo.comp_info[0].h_samp_factor, w, h);
  if(cinfo.jpeg_color_space==JCS_GRAYSCALE) {
    cinfo.out_color_space=JCS_GRAYSCALE;
    colorspace=GL_LUMINANCE;
    in_colorspace=GL_LUMINANCE;
  } else if(glext->use_yuv_complex&&cinfo.jpeg_color_space==JCS_YCbCr&&((w&0xf)==0)&&((h&0xf)==0)) {
    cinfo.raw_data_out=TRUE;
    read_whole=false;
    colorspace=PIXMAP_COLORSPACE_YUV_COMPLEX;
    in_colorspace=PIXMAP_COLORSPACE_YUV_COMPLEX;
  } else if(glext->use_yuv&&cinfo.jpeg_color_space==JCS_YCbCr) {
    cinfo.out_color_space=JCS_YCbCr;
    colorspace=PIXMAP_COLORSPACE_YUV;
    in_colorspace=PIXMAP_COLORSPACE_YUV_COMPLEX;
  } else {
    if(cinfo.jpeg_color_space==JCS_YCbCr)
      in_colorspace=PIXMAP_COLORSPACE_YUV_COMPLEX;
    else
      in_colorspace=-1;

    cinfo.out_color_space=JCS_RGB;
    colorspace=GL_RGB;
  }
  printf("input colorspace: %s, outcs: %s\n", get_color_space_string(in_colorspace),
      get_color_space_string(colorspace));
  jpeg_start_decompress(&cinfo);
  w=cinfo.output_width;
  h=cinfo.output_height;
  compute_alloc_size();
  if(colorspace==PIXMAP_COLORSPACE_YUV_COMPLEX) {
    wpadded=cinfo.comp_info[0].width_in_blocks*8;
    hpadded=cinfo.comp_info[0].height_in_blocks*8;
    iterations=cinfo.comp_info[0].height_in_blocks/cinfo.comp_info[0].v_samp_factor;
    int i;
    for(i=0; i<3; i++) {
      allocw_per_channel[i]=cinfo.comp_info[i].width_in_blocks*8;
      alloch_per_channel[i]=cinfo.comp_info[i].v_samp_factor*8;
    }
  }

  image_valid=true;
  use_data=read_whole;
  if(read_whole) {
    data=(unsigned char *)malloc(alloc_size);
    int row_stride=w*components;
    JSAMPROW row_pointer[1];
    while(cinfo.output_scanline<cinfo.output_height) {
      row_pointer[0]=&data[cinfo.output_scanline*row_stride];
      jpeg_read_scanlines(&cinfo, row_pointer, 1);
    }
    jpeg_finish_decompress(&cinfo);
    fclose(ImageFile);
    jpeg_destroy_decompress(&cinfo);
  }
}

void JPEGPixmap::read_scanlines(unsigned char **data[3])
{
  int max_lines=alloch_per_channel[0];
  //printf("data=%x, os=%d\n", data, cinfo.output_scanline);
  //printf("read scanlines, ml=%d\n", alloch_per_channel[0]);
  jpeg_read_raw_data(&cinfo, (JSAMPIMAGE) data, max_lines);
  //printf("done\n");
}

bool JPEGPixmap::read_scanline(char *data)
{
    int row_stride=w*bytespersample;
    if(cinfo.output_scanline<h) {
      jpeg_read_scanlines(&cinfo, (JSAMPROW *) &data, 1);
      return true;
    } else if(ImageFile) {
      jpeg_finish_decompress(&cinfo);
      jpeg_destroy_decompress(&cinfo);
      fclose(ImageFile);
      ImageFile=NULL;
      image_valid=false;
    }
    return false;
}
