#ifndef __PIXMAP_H
#define __PIXMAP_H

#include "conf.hh"
#include <stdio.h>

#define PIXMAP_COLORSPACE_BITMAP         100
#define PIXMAP_COLORSPACE_YUV            101
#define PIXMAP_COLORSPACE_YUV_COMPLEX    102

// A Pixmap represents an image not yet in the video memory.
// It can be either in the RAM or generated on-fly.
struct Pixmap {
    int refcnt; //reference counter
    // computes allocw, alloch, components and alloc_size from w, h, colorspace and the availability of
    // non-rectagular textures
    void compute_alloc_size(); 
  public:
    int w, h; // size of the image
    int components; // number of components
    int wpadded, hpadded; // padded width and height of image; usually w and h (even for rectangular textures)
    int alloc_size; // size of memory to be allocated (can be more than wpadded and hpadded because of a bug in Opengl drivers)

    // OpenGL colorspace or one of the additional ones:
    // GL_RGB (RGB, 24 bits), GL_LUMINANCE(grayscale), PIXMAP_COLORSPACE_YUV_COMPLEX (YUV, 24 bits),
    // GL_RGBA (RGB, 32 bits) or PIXMAP_COLORSPACE_YUV_COMPLEX (see read_scanlines for more info).
    int colorspace;

    bool flip; // should the image be flipped?
    bool use_data; // true->whole image is read to data, false->you have to read every scanline/basic unit
    bool image_valid; // after creating the Pixmap object, image_valid indicates that there was no error
    unsigned char *data; // data holds the pixmap data
    virtual bool read_scanline(char *data);

    // When using PIXMAP_COLORSPACE_YUV_COMPLEX, the color channels are used separately, you can get the
    // lines with the read_scanlines function. allocw_per_channel and alloch_per_channel shows the full
    // memory to be allocated per color channel, iterations is the number of read_scanlines calls
    // for the image to be completely read. You get alloch_per_channel[i]/iterations * allocw_per_channel[i]
    // bytes at each call. You must supply an unsigned char *[3] array.
    // For other color channels, the value of these variables and the read_scanlines function are undefined.
    int alloch_per_channel[3], allocw_per_channel[3], iterations; 
    virtual void read_scanlines(unsigned char **data[3]);

    Pixmap();
    virtual ~Pixmap();
    void ref();
    void unref();
};

class PixmapFactory
{
  public:
    virtual Pixmap *create(const char *fn);
    virtual const char *get_name();
    virtual bool accept_extension(const char *ext);
};


extern PixmapFactory *pixmap_factories[];



#endif /* __PIXMAP_H */
